import torch
import torch.nn as nn
import os
import torch.nn.functional as F
__all__ = ['MobileNetV2', 'mobilenet_v2']
#flag = [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0]
#flag = [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1]

class ConvBNReLU(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, groups=1):
        padding = (kernel_size - 1) // 2
        super(ConvBNReLU, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes),
            nn.ReLU6(inplace=True)
        )


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup
        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))
        layers.extend([
            # dw
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            # pw-linear
            nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
            nn.BatchNorm2d(oup),
        ])
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(self, num_classes=10, width_mult=1.0):
        super(MobileNetV2, self).__init__()
        block = InvertedResidual
        input_channel = 32
        last_channel = 1280
        
        ## CIFAR10
        inverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 1], # Stride 2 -> 1 for CIFAR-10
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]
        ## END

        # building first layer
        input_channel = int(input_channel * width_mult)
        self.last_channel = int(last_channel * max(1.0, width_mult))
        
        # CIFAR10: stride 2 -> 1
        features = [ConvBNReLU(3, input_channel, stride=1)]
        # END
        
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = int(c * width_mult)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(block(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        features.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        # building classifier
        self.classifier = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(self.last_channel, num_classes),
        )

        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def forward(self, x, linbp=False):
        if(linbp):
            flag = [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0]
        else:
            flag = [1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1]
        for i,blocks in enumerate(self.features):
            if flag[i] == 0 and isinstance(blocks,ConvBNReLU):
                for j,mn in enumerate(blocks):
                    if(isinstance(mn,nn.ReLU6)):
                        x_p1 = F.relu(-x)
                        x_p2 = F.relu(x-6)
                        x = x + x_p1.data + x_p2.data 
                    else:
                        x = mn(x)
            elif flag[i] == 0 and isinstance(blocks,InvertedResidual):
                use_res = blocks.use_res_connect
                temp = x
                for j,layer in enumerate(blocks.conv):
                    if(isinstance(layer,ConvBNReLU)):
                        for k,mn in enumerate(layer):
                            if(isinstance(mn,nn.ReLU6)):
                                x_p1 = F.relu(-x)
                                x_p2 = F.relu(x-6)
                                x = x + x_p1.data + x_p2.data
                            else:    
                                x = mn(x)
                    else:
                        x = layer(x)
                if(use_res):
                    x = temp + x
                else:
                    x = x
            else:
                x = blocks(x)

        #x = self.features(x)
        x = x.mean([2, 3])
        x = self.classifier(x)
        return x


def mobilenet_v2(pretrained=False, progress=True, device='cpu', **kwargs):
    """
    Constructs a MobileNetV2 architecture from
    `"MobileNetV2: Inverted Residuals and Linear Bottlenecks" <https://arxiv.org/abs/1801.04381>`_.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    model = MobileNetV2(**kwargs)
    if pretrained:
        script_dir = os.path.dirname(__file__)
        state_dict = torch.load(script_dir+'/state_dicts/mobilenet_v2.pt', map_location=device)
        model.load_state_dict(state_dict)
    return model
